"""
This is a modified version of source code from the Accerciser project
(http://live.gnome.org/accerciser).

Backend to the console plugin.

@author: Eitan Isaacson
@organization: IBM Corporation
@copyright: Copyright (c) 2007 IBM Corporation
@license: BSD

All rights reserved. This program and the accompanying materials are made
available under the terms of the BSD which accompanies this distribution, and
is available at U{http://www.opensource.org/licenses/bsd-license.php}

$Id: tkipshell.py 1037 2009-08-07 19:03:34Z jmiller $
"""

import re
import sys
import os
import subprocess

from StringIO import StringIO
import Tkinter, tkMessageBox, Pmw
import IPython

# eol's are saved in the history as \n
# eol's are recalled from the history as <<eol>> and re-entered as <<eol>>
#   by the client...  but <<eol>>'s are replaced with \n as they are read in.
# This let's a single line history function as a multi-line history.

EOL_HACK = "<<eol>>"
eol = re.compile("\n")
EOL_HACK_RE = re.compile(EOL_HACK)

NOTOUCH_TAG = "**notouch**"
INPUT_TAG = "**input**"
OUTPUT_TAG = "**otuput**"

def munge_eol(s):
    return eol.sub(EOL_HACK, s.rstrip("\n"))

# The seemingly nonsensical rstrip limits the string to a single EOL.
def unmunge_eol(s):
    t = EOL_HACK_RE.sub("\n", s)
    return t.rstrip("\n") + "\n"

def debug(*args, **keys):
        # return   # Remove this for console debug messages not trapped by IPython.
        sep = keys.get("sep", " ")
        eol  = keys.get("eol", "\n")
        for a in args:
                sys.__stdout__.write(str(a) + sep)
        sys.__stdout__.write(eol)

class IterableIPShell:
    def __init__(self,argv=None,user_ns=None,user_global_ns=None,
                 cin=None, cout=None,cerr=None, input_func=None):
        if input_func:
            IPython.iplib.raw_input_original = input_func
        if cin:
            IPython.Shell.Term.cin = cin
        if cout:
            IPython.Shell.Term.cout = cout
        if cerr:
            IPython.Shell.Term.cerr = cerr

        if argv is None:
            argv=[]

        # This is to get rid of the blockage that occurs during
        # IPython.Shell.InteractiveShell.user_setup()
        IPython.iplib.raw_input = lambda x: None

        self.term = IPython.genutils.IOTerm(cin=cin, cout=cout, cerr=cerr)
        os.environ['TERM'] = 'dumb'
        excepthook = sys.excepthook
        self.IP = IPython.Shell.make_IPython(argv,user_ns=user_ns,
                                             user_global_ns=user_global_ns, embedded=True,
                                             shell_class=IPython.Shell.InteractiveShell)
        sys.excepthook = excepthook
        self.iter_more = 0
        self._historyToEnd()
        self.complete_sep =  re.compile('[\s\{\}\[\]\(\)]')

    def execute(self):
        self._historyToEnd() 
        orig_stdout = sys.stdout
        sys.stdout = IPython.Shell.Term.cout
        try:
            line = self.IP.raw_input(None, self.iter_more)
            line = unmunge_eol(line)  # replace <<eol>> with real \n
            # if self.IP.autoindent:
            #  self.IP.readline_startup_hook(None)
        except KeyboardInterrupt:
            self.IP.write('\nKeyboardInterrupt\n')
            self.IP.resetbuffer()
            # keep cache in sync with the prompt counter:
            self.IP.outputcache.prompt_count -= 1
            if self.IP.autoindent:
                self.IP.indent_current_nsp = 0
            self.iter_more = 0
        except:
            self.IP.showtraceback()
            self.iter_more = 0
        else:
            self.iter_more = self.IP.push(line)
            if (self.IP.SyntaxTB.last_syntax_error and
                self.IP.rc.autoedit_syntax):
                self.IP.edit_syntax_error()
        if self.iter_more:
            self.prompt = "..." # str(self.IP.outputcache.prompt2)
            # if self.IP.autoindent:
            #  self.IP.readline_startup_hook(self.IP.pre_readline)
        else:
            self.prompt = str(self.IP.outputcache.prompt1)
        sys.stdout = orig_stdout

    def historyBack(self):
        if self.history_level != 1:
            self.history_level -= 1
        return self._getHistory()

    def historyForward(self):
        if self.history_level != 0:
            self.history_level += 1
        return self._getHistory()

    def _getHistory(self):
        iphist = self.IP.user_ns['In']
        if self.history_level >= len(iphist) :
            self._historyToEnd()
            return ''
        if self.history_level < 0:
            self.history_level = len(iphist)-1
        rv = iphist[self.history_level]
        return munge_eol(rv)

    def _historyToEnd(self):
        self.history_level = 0

    def updateNamespace(self, ns_dict):
        self.IP.user_ns.update(ns_dict)

    def complete(self, line):
        split_line = self.complete_sep.split(line)
        possibilities = self.IP.complete(split_line[-1])
        if possibilities:
            common_prefix = reduce(self._commonPrefix, possibilities)
            completed = line[:-len(split_line[-1])]+common_prefix
        else:
            completed = line
        return completed, possibilities

    def _commonPrefix(self, str1, str2):
        for i in range(len(str1)):
            if not str2.startswith(str1[:i+1]):
                return str1[:i]
        return str1

ansi_colors =  {'0;30': 'Black',
                '0;31': 'Red',
                '0;32': 'Green',
                '0;33': 'Brown',
                '0;34': 'Blue',
                '0;35': 'Purple',
                '0;36': 'Cyan',
                '0;37': 'LightGray',
                '1;30': 'DarkGray',
                '1;31': 'DarkRed',
                '1;32': 'SeaGreen',
                '1;33': 'Yellow',
                '1;34': 'LightBlue',
                '1;35': 'MediumPurple',
                '1;36': 'LightCyan',
                '1;37': 'White'}

ransi_colors = dict([(v.lower(),k) for (k,v) in ansi_colors.items()])

INPUT_TAG = "input"

class TkConsoleView(Pmw.ScrolledText):
    def __init__(self,root):
        Pmw.ScrolledText.__init__(self,root)

        # Keeps track of where the insert cursor should be on the entry line
        self.mark = 'scroll_mark'
        self.mark_set(self.mark,Tkinter.END)
        self.mark_gravity(self.mark,Tkinter.RIGHT)

        # Set the tags for colouring the text
        for code in ansi_colors:
            self.tag_config(code, foreground=ansi_colors[code])

        self.tag_config(NOTOUCH_TAG) # Tag for areas of widget that aren't editable

        # color_pat matches the color tags and places these in a
        # group match character with hex value 01 (start of heading?)
        # zero or more times, followed by the hex character 1b
        # (escape) then "[" and group ...things.. followed by m (?)
        # and then hex character 02 (start of text) zero or more times
        self.color_pat = re.compile('\x01?\x1b\[(.*?)m\x02?')

        self.line_start = 'line_start' # Tracks start of user input on the line (excluding prompt)
        self.mark_set(self.line_start,Tkinter.INSERT)
        self.mark_gravity(self.line_start,Tkinter.LEFT)
        self.current_input_start_index = Tkinter.INSERT
        self._setBindings()

    def colorize(self, text, color):
        return '\x01\x1b[%sm\x02%s' % (ransi_colors[color], text)

    def write(self, text, tags=[NOTOUCH_TAG]):
        """Insert text at END,  handling ANSI color escapes and adding
        the specified tags.
        """
        tags = tuple(tags)
        # Always write to end of buffer.
        self.cursor_to_prompt()
        segments = self.color_pat.split(text)
        # First is blank line
        segment = segments.pop(0)
        self.insert(Tkinter.END, segment, tags)
        if segments:
            # Just return the colour tags
            ansi_tags = self.color_pat.findall(text)
            for tag in ansi_tags:
                i = segments.index(tag)
                self.insert(Tkinter.END, segments[i+1], tags+(tag,))
                segments.pop(i)

    def showBanner(self,banner):
        """Print the supplied banner on starting the shell"""
        self.write(banner)

    def showPrompt(self, prompt):
        self.write(prompt)
        self.mark_set(self.line_start,Tkinter.INSERT)
 
    def changeLine(self, text):
        """Replace the input text at the prompt with the given text."""
        self.delete(self.line_start,"%s lineend" % self.line_start)
        self.write(text,  tags=[INPUT_TAG])
        self.cursor_to_prompt()

    def getCurrentLine(self):
        rv = self.get(self.line_start,Tkinter.END)
        self.tag_add(NOTOUCH_TAG,  self.line_start, Tkinter.END+"-1c")
        return rv

    def showReturned(self, text):
        nonempty = text.strip()
        text = self.colorize(text, "seagreen")
        if nonempty:
            self.write("\n"+text)
        self.write("\n", tags=[NOTOUCH_TAG, OUTPUT_TAG])
        self.showPrompt(self.prompt)

    def _setBindings(self):
        """ Bind the keys we require.  REM: if a bound function
        returns "break" then no other bindings are called If it
        returns None, then the other default bindings are called.
        """
        text = self.component("text")
        text.bind("<Key>",self.processKeyPress)
        text.bind("<Return>",self.processEnterPress)
        text.bind("<Up>",self.processUpPress)
        text.bind("<Down>",self.processDownPress)
        text.bind("<Tab>",self.processTabPress)
        text.bind("<BackSpace>",self.processBackSpacePress)
        text.bind("<Control-d>", self.quit)

    def ignore(self, e):
        return "break"

    def quit(self, *args):
        if tkMessageBox.askokcancel("Quit", "Really quit?"):
            sys.exit()
            
    def _indexInTagRanges(self, index, tag):
        ranges = self.tag_ranges(tag)
        for i in range(0,len(ranges),2):
            if (self.compare(index,'>=',ranges[i] ) and 
                 self.compare(index,'<=',ranges[i+1] )):
                        return (ranges[i], ranges[i+1])
        return None

    def isEditable(self):
        """ Scan the notouch tag range in pairs and see if the INSERT
        index falls between any of them.
        """
        range = self._indexInTagRanges(Tkinter.INSERT, NOTOUCH_TAG)
        if range is None:
                return True
        return range is None
    
    def processKeyPress(self,event):
        if not self.isEditable():
            # Move cursor mark to start of line
            self.mark_set(Tkinter.INSERT,self.mark)
        # Make sure line_start follows inserted text
        self.mark_set(self.mark,"%s+1c" % Tkinter.INSERT)

    def processBackSpacePress(self,event):
        if not self.isEditable():
            return "break"
        
    def _indexLine(self, mark):
        return int(self.index(mark).split(".")[0])
    
    def processEnterPress(self,event):
        insert_line = self._indexLine(Tkinter.INSERT)
        end_line = self._indexLine(Tkinter.END)
        if insert_line != end_line-1:
            self._enterNotAtPrompt()
        else:
            self.processLine()
        return "break" # Need break to stop the other bindings being called

    def processUpPress(self,event):
        self.cursor_to_prompt()
        self.changeLine(self.historyBack())
        return "break"

    def processDownPress(self,event):
        self.cursor_to_prompt()
        self.changeLine(self.historyForward())
        return "break"

    def processTabPress(self,event):
        if not self.getCurrentLine().strip():
            return
        completed, possibilities = self.complete(self.getCurrentLine())
        if len(possibilities) > 1:
            slice = self.getCurrentLine()
            self.write('\n')
            for symbol in possibilities:
                self.write(symbol+'\n')
            self.showPrompt(self.prompt)
        self.changeLine(completed or slice)
        return "break"
    
    def _enterNotAtPrompt(self):
        "Handle enter when the cursor is not at the end of the text."
        current_input = self._indexInTagRanges(Tkinter.INSERT, INPUT_TAG)
        if not current_input:  # Handle cursor in output text.
                self.cursor_to_prompt()
                return
        input = self.get(*current_input)
        self.changeLine(input)
        
    def clear(self):
        """Clears the edit window and prompts."""
        self.delete(1.0, Tkinter.END)
        self.mark_set(self.mark, Tkinter.END)
        self.mark_set(self.line_start, Tkinter.END)
        self.showPrompt(self.prompt)
        self.cursor_to_prompt()
        
    def _clickAtEnd(self):
        """Fake a mouse event to move the cursor to the end of the text
        widget.
        """
        text = self.component("text")
        try:
            end_x, end_y, width, height = text.bbox(Tkinter.END+"-1c")
            text.event_generate("<Button-1>", x=end_x, y=end_y)
        except:
            pass

    def cursor_to_prompt(self):
        """Moves the cursor to 1 character after the prompt which is
        assumed to be the last thing in the edit buffer.
        """
        self.tag_remove(Tkinter.SEL, "1.0", Tkinter.END)
        self.mark_set(Tkinter.INSERT, Tkinter.END+"-1c")
        # self.see(Tkinter.END)
        self.see(self.line_start)

class IPythonView(TkConsoleView, IterableIPShell):
    def __init__(self,root,banner=None, **keys):
        TkConsoleView.__init__(self,root)
        self.cout = StringIO()
        IterableIPShell.__init__(self, cout=self.cout,cerr=self.cout,
                                 input_func=self.raw_input, **keys)
        if banner:
            self.showBanner(banner)
        self.execute()
        self.cout.truncate(0)
        self.showPrompt(self.prompt)
        self.interrupt = False

    def raw_input(self, prompt=''):
        if hasattr(self, "interrupt") and self.interrupt:
            self.interrupt = False
            raise KeyboardInterrupt
        return self.getCurrentLine()

    def processLine(self):
        self.execute()
        rv = self.cout.getvalue()
        if rv: 
                rv = rv.strip('\n')
        self.showReturned(rv)
        self.cout.truncate(0)
        self.cursor_to_prompt()

# Support for launching subprocesses,  currently unused.
class SubprocException(Exception):
    def __init__(self, out, err, code):
        self.out = out
        self.err = err
        self.code = code

def execute(command, use_exceptions=False,  input=None, cwd=None):
    """Execute command as a subprocess.  Returns  (code, stdout, stderr) """
    if not isinstance(command, str):
        debug("command:", repr(command))
        cmd = " ".join(command)
    else:
        cmd = command

    debug("executing: ", repr(cmd))
    
    proc = subprocess.Popen(
        command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, stdin=subprocess.PIPE,
        cwd=cwd)
    
    if input is not None:
        debug("stdin:",  repr(input))
        proc.stdin.write(input)
    
    out, err = proc.communicate()
    code = proc.wait()
    
    debug("stdout:", out)
    debug("stderr:", err)
    debug("code:", code)

    if use_exceptions and code != 0:
        raise SubprocException(out,err,code)

    return (code, out, err)

if __name__ == "__main__":
    root = Tkinter.Tk()
    s=IPythonView(root)
    s.pack()
    root.mainloop()
